#
# This software is Copyright ©️ 2020 The University of Southern California. All Rights Reserved.
# Permission to use, copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice and subject to the full license file found in the root of this software deliverable. Permission to make commercial use of this software may be obtained by contacting:  USC Stevens Center for Innovation University of Southern California 1150 S. Olive Street, Suite 2300, Los Angeles, CA 90115, USA Email: accounting@stevens.usc.edu
#
# The full terms of this copyright and license should always be found in the root directory of this software deliverable as "license.txt" and if these terms are not found with this software, please contact the USC Stevens Center for the full license.
#
import pytest
from unittest.mock import patch

from transcribe import (
    TranscribeBatchResult,
    TranscribeJob,
    TranscribeJobRequest,
    TranscribeJobStatus,
    TranscribeJobsUpdate,
)

from .helpers import (
    run_transcribe_test,
    AwsTranscribeGetJobCall,
    AwsTranscribeListJobsCall,
    AwsTranscribeStartJobCall,
    TranscribeTestFixture,
    TEST_AWS_REGION,
    TEST_TRANSCRIBE_SOURCE_BUCKET,
)


@patch("boto3.client")
@pytest.mark.parametrize(
    "fixture",
    [
        (
            TranscribeTestFixture(
                requests=[
                    TranscribeJobRequest(jobId="m1-u1", sourceFile="/audio/m1/u1.wav")
                ],
                override_expected_start_job_calls=[
                    AwsTranscribeStartJobCall(
                        expected_args={
                            "TranscriptionJobName": "m1-u1",
                            "LanguageCode": "en-US",
                            "Media": {
                                "MediaFileUri": f"https://s3.{TEST_AWS_REGION}.amazonaws.com/{TEST_TRANSCRIBE_SOURCE_BUCKET}/m1-u1.wav"
                            },
                            "MediaFormat": "wav",
                        }
                    )
                ],
                list_jobs_calls=[
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "m1-u1",
                                    "TranscriptionJobStatus": "COMPLETED",
                                }
                            ]
                        }
                    )
                ],
                get_job_calls=[
                    AwsTranscribeGetJobCall(
                        name="m1-u1",
                        result={
                            "TranscriptionJob": {
                                "TranscriptionJobStatus": "COMPLETED",
                                "Transcript": {
                                    "TranscriptFileUri": "http://fake/m1-u1"
                                },
                            }
                        },
                        transcribe_url_response={
                            "results": {
                                "transcripts": [{"transcript": ""}],
                                "items": [],
                            },
                            "status": "COMPLETED",
                        },
                    )
                ],
                expected_sleep_calls=[],
                expected_result=TranscribeBatchResult(
                    transcribeJobsById={
                        "m1-u1": TranscribeJob(
                            jobId="m1-u1",
                            sourceFile="/audio/m1/u1.wav",
                            mediaFormat="wav",
                            status=TranscribeJobStatus.SUCCEEDED,
                            transcript="",
                        )
                    }
                ),
                expected_on_update_calls=[
                    TranscribeJobsUpdate(
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "m1-u1": TranscribeJob(
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.SUCCEEDED,
                                    transcript="",
                                )
                            }
                        ),
                        idsUpdated=["m1-u1"],
                    )
                ],
            )
        )
    ],
)
def test_it_accepts_a_completed_result_with_empty_transcript(
    mock_boto3_client, fixture: TranscribeTestFixture
):
    run_transcribe_test(mock_boto3_client, fixture)



#
# This software is Copyright ©️ 2020 The University of Southern California. All Rights Reserved.
# Permission to use, copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice and subject to the full license file found in the root of this software deliverable. Permission to make commercial use of this software may be obtained by contacting:  USC Stevens Center for Innovation University of Southern California 1150 S. Olive Street, Suite 2300, Los Angeles, CA 90115, USA Email: accounting@stevens.usc.edu
#
# The full terms of this copyright and license should always be found in the root directory of this software deliverable as "license.txt" and if these terms are not found with this software, please contact the USC Stevens Center for the full license.
#
from unittest.mock import patch
from transcribe import init_transcription_service, TranscriptionService
from .helpers import TEST_SERVICE_CONFIG


@patch("boto3.client")
def test_it_creates_a_service_instance_for_module_path(mock_boto3_client):
    service = init_transcription_service(
        module_path="transcribe_aws", config=TEST_SERVICE_CONFIG
    )
    assert isinstance(service, TranscriptionService)
    mock_boto3_client.assert_any_call(
        "s3",
        region_name="fake-region",
        aws_access_key_id="fake-access-key-id",
        aws_secret_access_key="fake-secret-access-key",
    )
    mock_boto3_client.assert_any_call(
        "transcribe",
        region_name="fake-region",
        aws_access_key_id="fake-access-key-id",
        aws_secret_access_key="fake-secret-access-key",
    )



#
# This software is Copyright ©️ 2020 The University of Southern California. All Rights Reserved.
# Permission to use, copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice and subject to the full license file found in the root of this software deliverable. Permission to make commercial use of this software may be obtained by contacting:  USC Stevens Center for Innovation University of Southern California 1150 S. Olive Street, Suite 2300, Los Angeles, CA 90115, USA Email: accounting@stevens.usc.edu
#
# The full terms of this copyright and license should always be found in the root directory of this software deliverable as "license.txt" and if these terms are not found with this software, please contact the USC Stevens Center for the full license.
#
import pytest
from unittest.mock import patch

from transcribe import (
    TranscribeBatchResult,
    TranscribeJob,
    TranscribeJobRequest,
    TranscribeJobStatus,
    TranscribeJobsUpdate,
)

from transcribe_aws import DEFAULT_POLL_INTERVAL

from .helpers import (
    run_transcribe_test,
    AwsTranscribeGetJobCall,
    AwsTranscribeListJobsCall,
    TranscribeTestFixture,
)


@patch("boto3.client")
@pytest.mark.parametrize(
    "fixture",
    [
        (
            TranscribeTestFixture(
                batch_id="b1",
                requests=[
                    TranscribeJobRequest(jobId="m1-u1", sourceFile="/audio/m1/u1.wav"),
                    TranscribeJobRequest(jobId="m1-u2", sourceFile="/audio/m1/u2.wav"),
                    TranscribeJobRequest(jobId="m1-u3", sourceFile="/audio/m1/u3.wav"),
                ],
                list_jobs_calls=[
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "QUEUED",
                                }
                            ]
                        }
                    ),
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u2",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                            ]
                        }
                    ),
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u2",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u3",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                            ]
                        }
                    ),
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "FAILED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u2",
                                    "TranscriptionJobStatus": "IN_PROGRESS",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u3",
                                    "TranscriptionJobStatus": "QUEUED",
                                },
                            ]
                        }
                    ),
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "FAILED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u2",
                                    "TranscriptionJobStatus": "FAILED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u3",
                                    "TranscriptionJobStatus": "IN_PROGRESS",
                                },
                            ]
                        }
                    ),
                    AwsTranscribeListJobsCall(
                        result={
                            "TranscriptionJobSummaries": [
                                {
                                    "TranscriptionJobName": "b1-m1-u1",
                                    "TranscriptionJobStatus": "FAILED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u2",
                                    "TranscriptionJobStatus": "FAILED",
                                },
                                {
                                    "TranscriptionJobName": "b1-m1-u3",
                                    "TranscriptionJobStatus": "COMPLETED",
                                },
                            ]
                        }
                    ),
                ],
                get_job_calls=[
                    AwsTranscribeGetJobCall(
                        name="b1-m1-u3",
                        result={
                            "TranscriptionJob": {
                                "TranscriptionJobStatus": "COMPLETED",
                                "Transcript": {
                                    "TranscriptFileUri": "http://fake/b1-m1-u3"
                                },
                            }
                        },
                        transcribe_url_response={
                            "results": {
                                "transcripts": [
                                    {
                                        "transcript": "transcript for mentor m1 and utterance u3 even though other jobs failed"
                                    }
                                ]
                            }
                        },
                    )
                ],
                expected_sleep_calls=[
                    DEFAULT_POLL_INTERVAL,
                    DEFAULT_POLL_INTERVAL,
                    DEFAULT_POLL_INTERVAL,
                ],
                expected_result=TranscribeBatchResult(
                    transcribeJobsById={
                        "b1-m1-u1": TranscribeJob(
                            batchId="b1",
                            jobId="m1-u1",
                            sourceFile="/audio/m1/u1.wav",
                            mediaFormat="wav",
                            status=TranscribeJobStatus.FAILED,
                        ),
                        "b1-m1-u2": TranscribeJob(
                            batchId="b1",
                            jobId="m1-u2",
                            sourceFile="/audio/m1/u2.wav",
                            mediaFormat="wav",
                            status=TranscribeJobStatus.FAILED,
                        ),
                        "b1-m1-u3": TranscribeJob(
                            batchId="b1",
                            jobId="m1-u3",
                            sourceFile="/audio/m1/u3.wav",
                            mediaFormat="wav",
                            status=TranscribeJobStatus.SUCCEEDED,
                            transcript="transcript for mentor m1 and utterance u3 even though other jobs failed",
                        ),
                    }
                ),
                expected_on_update_calls=[
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u1"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.UPLOADED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u1"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u2"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.UPLOADED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u2"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.NONE,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u3"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.UPLOADED,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u3"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u1", "b1-m1-u2"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.FAILED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.IN_PROGRESS,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.QUEUED,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u2", "b1-m1-u3"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.FAILED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.FAILED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.IN_PROGRESS,
                                ),
                            }
                        ),
                    ),
                    TranscribeJobsUpdate(
                        idsUpdated=["b1-m1-u3"],
                        result=TranscribeBatchResult(
                            transcribeJobsById={
                                "b1-m1-u1": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u1",
                                    sourceFile="/audio/m1/u1.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.FAILED,
                                ),
                                "b1-m1-u2": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u2",
                                    sourceFile="/audio/m1/u2.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.FAILED,
                                ),
                                "b1-m1-u3": TranscribeJob(
                                    batchId="b1",
                                    jobId="m1-u3",
                                    sourceFile="/audio/m1/u3.wav",
                                    mediaFormat="wav",
                                    status=TranscribeJobStatus.SUCCEEDED,
                                    transcript="transcript for mentor m1 and utterance u3 even though other jobs failed",
                                ),
                            }
                        ),
                    ),
                ],
            )
        )
    ],
)
def test_it_continues_to_completion_when_some_jobs_fail(
    mock_boto3_client, fixture: TranscribeTestFixture
):
    run_transcribe_test(mock_boto3_client, fixture)



#
# This software is Copyright ©️ 2020 The University of Southern California. All Rights Reserved.
# Permission to use, copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice and subject to the full license file found in the root of this software deliverable. Permission to make commercial use of this software may be obtained by contacting:  USC Stevens Center for Innovation University of Southern California 1150 S. Olive Street, Suite 2300, Los Angeles, CA 90115, USA Email: accounting@stevens.usc.edu
#
# The full terms of this copyright and license should always be found in the root directory of this software deliverable as "license.txt" and if these terms are not found with this software, please contact the USC Stevens Center for the full license.
#
import logging
import requests
import os
import re
from typing import Any, Callable, Dict, Iterable, List, Optional, Union
import time

import boto3

from botocore.exceptions import ClientError
from boto3_type_annotations.s3 import Client as S3Client
from boto3_type_annotations.transcribe import Client as TranscribeClient

from transcribe import (
    copy_shallow,
    requests_to_job_batch,
    register_transcription_service_factory,
    TranscribeBatchResult,
    TranscribeJob,
    TranscribeJobRequest,
    TranscribeJobsUpdate,
    TranscribeJobStatus,
    TranscriptionService,
)

_TRANSCRIBE_JOB_STATUS_BY_AWS_STATUS: Dict[str, TranscribeJobStatus] = {
    "QUEUED": TranscribeJobStatus.QUEUED,
    "IN_PROGRESS": TranscribeJobStatus.IN_PROGRESS,
    "FAILED": TranscribeJobStatus.FAILED,
    "COMPLETED": TranscribeJobStatus.SUCCEEDED,
}

DEFAULT_POLL_INTERVAL: float = 5.0


def _require_env(n: Union[str, List[str]], v: str = "") -> str:
    if v:
        return v
    env_names: List[str] = [n] if isinstance(n, str) else n
    for k in env_names:
        env_val = os.environ.get(k, "")
        if env_val:
            return env_val
    raise EnvironmentError(f"missing required env var {'|'.join(env_names)}")


def _prefix_require_env(n: str, v: str = "") -> str:
    return _require_env([f"TRANSCRIBE_{n}", n], v)


def _create_s3_client(
    aws_access_key_id: str = "", aws_secret_access_key: str = "", aws_region: str = ""
) -> S3Client:
    return boto3.client(
        "s3",
        region_name=_prefix_require_env("AWS_REGION", aws_region),
        aws_access_key_id=_prefix_require_env("AWS_ACCESS_KEY_ID", aws_access_key_id),
        aws_secret_access_key=_prefix_require_env(
            "AWS_SECRET_ACCESS_KEY",
            aws_secret_access_key,
        ),
    )


def _create_transcribe_client(
    aws_access_key_id: str = "", aws_secret_access_key: str = "", aws_region: str = ""
) -> TranscribeClient:
    return boto3.client(
        "transcribe",
        region_name=_prefix_require_env("AWS_REGION", aws_region),
        aws_access_key_id=_prefix_require_env("AWS_ACCESS_KEY_ID", aws_access_key_id),
        aws_secret_access_key=_prefix_require_env(
            "AWS_SECRET_ACCESS_KEY", aws_secret_access_key
        ),
    )


def _parse_aws_status(
    aws_status: str, default_status: TranscribeJobStatus = TranscribeJobStatus.NONE
) -> TranscribeJobStatus:
    return _TRANSCRIBE_JOB_STATUS_BY_AWS_STATUS.get(aws_status, default_status)


def _s3_file_exists(s3: S3Client, bucket: str, key: str) -> bool:
    try:
        s3.head_object(Bucket=bucket, Key=key)
    except ClientError as e:
        logging.error(e)
        return int(e.response["Error"]["Code"]) != 404
    return True


class AWSTranscriptionService(TranscriptionService):
    def _get_batch_status(self, batch_id: str) -> List[Dict[str, Any]]:
        result: List[Dict[str, Any]] = []
        cur_result_page = self.transcribe_client.list_transcription_jobs(
            JobNameContains=batch_id
        )
        while True:
            cur_result_summaries: List[Dict[str, Any]] = cur_result_page.get(
                "TranscriptionJobSummaries"
            )
            for r in cur_result_summaries:
                result.append(r)
            next_token = cur_result_page.get("NextToken", "")
            if not next_token:
                break
            cur_result_page = self.transcribe_client.list_transcription_jobs(
                JobNameContains=batch_id, NextToken=next_token
            )
        return result

    def _load_transcript(self, aws_job_name: str) -> str:
        aws_job = self.transcribe_client.get_transcription_job(
            TranscriptionJobName=aws_job_name
        )
        url = (
            aws_job.get("TranscriptionJob", {})
            .get("Transcript", {})
            .get("TranscriptFileUri", "")
        )
        if not url:
            raise Exception(f"unable to parse url for job '{aws_job_name}': {aws_job}")
        transcript_res = requests.get(url)
        transcript_res.raise_for_status()
        transcript_json = transcript_res.json()
        try:
            return transcript_json["results"]["transcripts"][0]["transcript"]
        except Exception:
            raise Exception(
                f"unable to parse transcript for job '{aws_job_name} and url {url}': {transcript_json}"
            )

    def get_s3_path(self, source_file: str, id: str) -> str:
        file_name = f"{id.lower()}{os.path.splitext(source_file)[1]}"
        return f"{self.s3_root_path}/{file_name}" if self.s3_root_path else file_name

    def init_service(self, config: Dict[str, Any] = {}, **kwargs):
        self.aws_region = config.get("AWS_REGION") or _prefix_require_env("AWS_REGION")
        self.s3_bucket_source = config.get(
            "TRANSCRIBE_AWS_S3_BUCKET_SOURCE"
        ) or _require_env("TRANSCRIBE_AWS_S3_BUCKET_SOURCE")
        self.s3_root_path = config.get(
            "TRANSCRIBE_AWS_S3_ROOT_PATH",
            os.environ.get("TRANSCRIBE_AWS_S3_ROOT_PATH", ""),
        )
        aws_access_key_id = config.get("AWS_ACCESS_KEY_ID") or _prefix_require_env(
            "AWS_ACCESS_KEY_ID"
        )
        aws_secret_access_key = config.get(
            "AWS_SECRET_ACCESS_KEY"
        ) or _prefix_require_env("AWS_SECRET_ACCESS_KEY")
        self.s3_client = _create_s3_client(
            aws_region=self.aws_region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.transcribe_client = _create_transcribe_client(
            aws_region=self.aws_region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.poll_interval = config.get("POLL_INTERVAL", DEFAULT_POLL_INTERVAL)

    def transcribe(
        self,
        transcribe_requests: Iterable[TranscribeJobRequest],
        batch_id: str = "",
        on_update: Optional[Callable[[TranscribeJobsUpdate], None]] = None,
        **kwargs,
    ) -> TranscribeBatchResult:
        logging.info(f"transcribe: assigning batch id {batch_id} to all jobs")
        result = TranscribeBatchResult(
            transcribeJobsById={
                j.get_fq_id(): j
                for j in requests_to_job_batch(batch_id, transcribe_requests)
            }
        )
        for i, job in enumerate(result.jobs()):
            result = self._upload_one(job, i, result, on_update)
            result = self._try_ensure_all_jobs_started(result, batch_id, on_update)
        while result.has_any_unresolved():
            if self.poll_interval > 0:
                time.sleep(self.poll_interval)
            result = self._try_ensure_all_jobs_started(result, batch_id, on_update)
            result = self._update_status(result, batch_id, on_update)
        return result

    def _send_on_update(
        self,
        result: TranscribeBatchResult,
        ids_updated: List[str],
        on_update: Optional[Callable[[TranscribeJobsUpdate], None]],
    ):
        if on_update and len(ids_updated) > 0:
            assert on_update is not None
            try:
                on_update(
                    TranscribeJobsUpdate(result=result, idsUpdated=sorted(ids_updated))
                )
            except Exception as ex:
                logging.exception(f"update handler raise exception: {ex}")
        return result

    def _try_ensure_all_jobs_started(
        self,
        result: TranscribeBatchResult,
        batch_id: str,
        on_update: Optional[Callable[[TranscribeJobsUpdate], None]],
    ):
        if not any(j.status == TranscribeJobStatus.UPLOADED for j in result.jobs()):
            return result
        result = copy_shallow(result)
        job_ids_started = []
        try:
            for job in result.jobs():
                if job.status != TranscribeJobStatus.UPLOADED:
                    continue
                jid = job.get_fq_id()
                item_s3_path = self.get_s3_path(job.sourceFile, jid)
                self.transcribe_client.start_transcription_job(
                    TranscriptionJobName=jid,
                    LanguageCode=job.languageCode,
                    Media={
                        "MediaFileUri": f"https://s3.{self.aws_region}.amazonaws.com/{self.s3_bucket_source}/{item_s3_path}"
                    },
                    MediaFormat=job.mediaFormat,
                )
                result.update_job(jid, status=TranscribeJobStatus.QUEUED)
                job_ids_started.append(jid)
        except BaseException as ex:
            if re.search("limitexceeded", str(ex), re.IGNORECASE):
                logging.info(
                    "received a limit-exceeded response from aws. Will try again to start this job shortly"
                )
            else:
                logging.exception(f"exception on start jobs: {ex}")
        if job_ids_started:
            self._send_on_update(result, job_ids_started, on_update)
        return result

    def _update_status(
        self,
        result: TranscribeBatchResult,
        batch_id: str,
        on_update: Optional[Callable[[TranscribeJobsUpdate], None]],
    ) -> TranscribeBatchResult:
        job_updates = self._get_batch_status(batch_id)
        ids_updated: List[str] = []
        result = copy_shallow(result)
        for ju in job_updates:
            try:
                jid = ju.get("TranscriptionJobName", "")
                jstatus = _parse_aws_status(
                    ju.get("TranscriptionJobStatus", ""),
                    default_status=TranscribeJobStatus.NONE,
                )
                if jstatus == TranscribeJobStatus.NONE:
                    raise ValueError(
                        f"job status has unknown value of {ju.get('TranscriptionJobStatus')}"
                    )
                if result.job_completed(jid, jstatus):
                    continue
                transcript = (
                    self._load_transcript(jid)
                    if jstatus == TranscribeJobStatus.SUCCEEDED
                    else ""
                )
                if result.update_job(jid, status=jstatus, transcript=transcript):
                    ids_updated.append(jid)
            except Exception as ex:
                logging.exception(f"failed to handle update for {ju}: {ex}")
        summary = result.summary()
        logging.info(
            f"transcribe [{summary.get_count_completed()}/{summary.get_count_total()}] completed. Statuses [SUCCEEDED: {summary.get_count(TranscribeJobStatus.SUCCEEDED)}, FAILED: {summary.get_count(TranscribeJobStatus.FAILED)}, QUEUED: {summary.get_count(TranscribeJobStatus.QUEUED)}, IN_PROGRESS: {summary.get_count(TranscribeJobStatus.IN_PROGRESS)}]."
        )
        self._send_on_update(result, ids_updated, on_update)
        return result

    def _upload_one(
        self,
        job: TranscribeJob,
        job_index: int,
        result: TranscribeBatchResult,
        on_update: Optional[Callable[[TranscribeJobsUpdate], None]],
    ) -> TranscribeBatchResult:
        jid = job.get_fq_id()
        result = copy_shallow(result)
        item_s3_path = self.get_s3_path(job.sourceFile, jid)
        logging.info(
            f"transcribe [{job_index + 1}/{len(result.transcribeJobsById)}] uploading audio to s3 bucket {self.s3_bucket_source} and path {item_s3_path}"
        )
        self.s3_client.upload_file(
            job.sourceFile,
            self.s3_bucket_source,
            item_s3_path,
            ExtraArgs={"ACL": "public-read"},
        )
        result = copy_shallow(result)
        result.update_job(jid, status=TranscribeJobStatus.UPLOADED)
        self._send_on_update(result, [jid], on_update)
        return result


register_transcription_service_factory("transcribe_aws", AWSTranscriptionService)



from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def _read_dependencies():
    requirements_file = "requirements.txt"
    with open(requirements_file) as fin:
        return [line.strip() for line in fin if line]


packages = find_packages()
requirements = _read_dependencies()


def _read_version():
    with open('VERSION') as version_file:
        return version_file.read().strip()


setup(
    name="py_transcribe_aws",
    version=_read_version(),
    author_email="larrykirschner@gmail.com",
    description="framework for synchronous batch speech-to-text transcription using backends like AWS, Watson, etc.",
    packages=packages,
    package_dir={'transcribe_aws': 'transcribe_aws'},
    package_data={
        "transcribe_aws": ["py.typed"],
    },
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type='text/markdown',
)



#
# This software is Copyright ©️ 2020 The University of Southern California. All Rights Reserved.
# Permission to use, copy, modify, and distribute this software and its documentation for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice and subject to the full license file found in the root of this software deliverable. Permission to make commercial use of this software may be obtained by contacting:  USC Stevens Center for Innovation University of Southern California 1150 S. Olive Street, Suite 2300, Los Angeles, CA 90115, USA Email: accounting@stevens.usc.edu
#
# The full terms of this copyright and license should always be found in the root directory of this software deliverable as "license.txt" and if these terms are not found with this software, please contact the USC Stevens Center for the full license.
#
from dataclasses import dataclass, field
import requests_mock
from typing import Any, Callable, Dict, List, Optional, Tuple
from unittest.mock import call, patch, Mock

from transcribe import TranscribeBatchResult, TranscribeJobRequest, TranscribeJobsUpdate
from transcribe_aws import AWSTranscriptionService

from tests.helpers import Bunch


class FakeLimitExceededException(BaseException):
    def __init__(self):
        super().__init__("LimitExceeded (fake)")


def raise_limit_exceeded():
    raise FakeLimitExceededException()


@dataclass
class AwsTranscribeStartJobCall:
    expected_args: Dict[str, Any]
    side_effect: Optional[Callable[[], None]] = None


@dataclass
class AwsTranscribeGetJobCall:
    name: str
    result: Dict[str, Any] = field(default_factory=lambda: {})
    transcribe_url_response: Dict[str, str] = field(default_factory=lambda: {})

    def is_success_result(self) -> bool:
        return (
            self.result
            and "TranscriptionJob" in self.result
            and self.result["TranscriptionJob"].get("TranscriptionJobStatus")
            == "COMPLETED"
        )

    def get_transcription_url(self) -> bool:
        return self.result["TranscriptionJob"]["Transcript"]["TranscriptFileUri"]


@dataclass
class AwsTranscribeListJobsCall:
    next_token: str = ""
    result: Dict[str, Any] = field(default_factory=lambda: {})


@dataclass
class TranscribeTestFixture:
    batch_id: str = ""
    requests: List[TranscribeJobRequest] = field(default_factory=lambda: [])
    get_job_calls: List[AwsTranscribeGetJobCall] = field(default_factory=lambda: [])
    list_jobs_calls: List[AwsTranscribeListJobsCall] = field(default_factory=lambda: [])
    override_expected_start_job_calls: List[AwsTranscribeStartJobCall] = field(
        default_factory=lambda: []
    )
    expected_on_update_calls: List[TranscribeJobsUpdate] = field(
        default_factory=lambda: []
    )
    expected_result: TranscribeBatchResult = field(
        default_factory=lambda: TranscribeBatchResult()
    )
    expected_sleep_calls: List[float] = field(default_factory=lambda: [])


TEST_TRANSCRIBE_SOURCE_BUCKET = "test_transcribe_source_bucket"
TEST_AWS_REGION = "fake_aws_region"


def create_service(mock_boto3_client) -> Tuple[AWSTranscriptionService, Any, Any]:
    mock_s3_client = Bunch(upload_file=Mock())
    mock_transcribe_client = Bunch(
        get_transcription_job=Mock(),
        list_transcription_jobs=Mock(),
        start_transcription_job=Mock(),
    )

    def return_clients(client_type, **kwargs):
        return (
            mock_s3_client
            if client_type == "s3"
            else mock_transcribe_client
            if client_type == "transcribe"
            else None
        )

    mock_boto3_client.side_effect = return_clients
    service = AWSTranscriptionService()
    service.init_service(
        config={
            "AWS_REGION": TEST_AWS_REGION,
            "AWS_SECRET_ACCESS_KEY": "fake_aws_secret_access_key",
            "AWS_ACCESS_KEY_ID": "fake_aws_access_key_id",
            "TRANSCRIBE_AWS_S3_BUCKET_SOURCE": TEST_TRANSCRIBE_SOURCE_BUCKET,
        }
    )
    return (service, mock_s3_client, mock_transcribe_client)


def run_transcribe_test(mock_boto3_client, fixture: TranscribeTestFixture):
    with patch("time.sleep") as mock_sleep:
        transcribe_service, mock_s3_client, mock_transcribe_client = create_service(
            mock_boto3_client
        )
        batch_id = fixture.batch_id
        spy_on_update = Mock()
        expected_upload_file_calls = []
        expected_start_transcription_job_calls = (
            [call(**c.expected_args) for c in fixture.override_expected_start_job_calls]
            if fixture.override_expected_start_job_calls
            else []
        )
        if fixture.override_expected_start_job_calls:
            start_transcription_job_side_effects = [
                sjc.side_effect for sjc in fixture.override_expected_start_job_calls
            ]

            def _side_effect(*arg, **kwargs):
                if start_transcription_job_side_effects:
                    e = start_transcription_job_side_effects.pop(0)
                    if e:
                        e()

            mock_transcribe_client.start_transcription_job.side_effect = _side_effect
        for r in fixture.requests:
            fqid = f"{batch_id}-{r.jobId}" if batch_id else r.jobId
            input_s3_path = transcribe_service.get_s3_path(r.sourceFile, fqid)
            expected_upload_file_calls.append(
                call(
                    r.sourceFile,
                    TEST_TRANSCRIBE_SOURCE_BUCKET,
                    input_s3_path,
                    ExtraArgs={"ACL": "public-read"},
                )
            )
            if not fixture.override_expected_start_job_calls:
                expected_start_transcription_job_calls.append(
                    call(
                        TranscriptionJobName=fqid,
                        Media={
                            "MediaFileUri": f"https://s3.{TEST_AWS_REGION}.amazonaws.com/{TEST_TRANSCRIBE_SOURCE_BUCKET}/{input_s3_path}"
                        },
                        MediaFormat=r.get_media_format(),
                        LanguageCode=r.get_language_code(),
                    )
                )
        expected_list_jobs_calls = []
        mock_list_call_responses = []
        for c in fixture.list_jobs_calls:
            expected_list_jobs_calls.append(
                call(JobNameContains=batch_id, NextToken=c.next_token)
            )
            mock_list_call_responses.append(c.result)
        mock_transcribe_client.list_transcription_jobs.side_effect = (
            mock_list_call_responses
        )
        expected_get_job_calls = []
        mock_get_job_responses = []
        with requests_mock.Mocker() as mock_requests:
            for c in fixture.get_job_calls:
                expected_get_job_calls.append(call(TranscriptionJobName=c.name))
                if c.is_success_result():
                    t_url = c.get_transcription_url()
                    mock_requests.get(t_url, json=c.transcribe_url_response)
                mock_get_job_responses.append(c.result)
            mock_transcribe_client.get_transcription_job.side_effect = (
                mock_get_job_responses
            )
            result = transcribe_service.transcribe(
                fixture.requests, on_update=spy_on_update, batch_id=batch_id
            )
            mock_s3_client.upload_file.assert_has_calls(expected_upload_file_calls)
            mock_transcribe_client.start_transcription_job.assert_has_calls(
                expected_start_transcription_job_calls
            )
            if fixture.expected_sleep_calls:
                mock_sleep.assert_has_calls(
                    [
                        call(sleep_interval)
                        for sleep_interval in fixture.expected_sleep_calls
                    ]
                )
            assert result.to_dict() == fixture.expected_result.to_dict()
            if fixture.expected_on_update_calls:
                expected_on_update_calls = [
                    call(u) for u in fixture.expected_on_update_calls
                ]
                spy_on_update.assert_has_calls(expected_on_update_calls)
